/*
 * TCSS 305Winter 2014
 * Assignment 4-SnapShop
 */

package snapshop.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import snapshop.filters.EdgeDetectFilter;
import snapshop.filters.EdgeHighlightFilter;
import snapshop.filters.Filter;
import snapshop.filters.FlipHorizontalFilter;
import snapshop.filters.FlipVerticalFilter;
import snapshop.filters.GrayscaleFilter;
import snapshop.filters.SharpenFilter;
import snapshop.filters.SoftenFilter;
import snapshop.image.Pixel;
import snapshop.image.PixelImage;

/**
 * The SnapShopGUI class extends JFrame and allows a JFrame object to be instantiated.  
 * This class allows a user to upload an image and apply various kinds of filters to it.  
 * It also lets the user save the image to a directory of choice.
 * 
 * @author Brian Dennis
 * @version 6 February 2014
 */
@SuppressWarnings("serial")
public class SnapShopGUI extends JFrame {

    /** A ToolKit for getting screen dimensions. */
    private static final Toolkit KIT = Toolkit.getDefaultToolkit();

    /** The Dimension of the screen. */
    private static final Dimension SCREEN_SIZE = KIT.getScreenSize();

    /** The width of the screen. */
    private static final int SCREEN_WIDTH = SCREEN_SIZE.width;

    /** The height of the screen. */
    private static final int SCREEN_HEIGHT = SCREEN_SIZE.height;
    
    /** The JFileChooser object used to open and save files. */
    private final JFileChooser myFileChooser;
    
    /** The JLabel object used to store a PixelImage as it's icon. */
    private final JLabel myLabelForViewingIcon;
    
    /** The PixelImage object representing the chosen picture. */
    private PixelImage myImage;
    
    /** A List object used to store JButtons for the purpose of enabling the buttons. */
    private final List<JButton> myButtonArray;
    
    /** An ArrayDeque object used to store 2 dimensional arrays of Pixel objects. 
        This stack is used to store previous Pixel versions of a filtered PixelImage. */
    private final ArrayDeque<Pixel[][]> myPixelStack;
    
    /** A string representing "Open.."  that I use to silence check style warnings. It is used 
        in places where I needed to use this string in two instances. */
    private final String myOpen;
    
    /** A string representing "Undo"  that I use to silence check style warnings. It is used 
    in places where I needed to use this string in two instances. */
    private final String myUndo;
    
    /**
     * The constructor of this class which calls the parent class, setting the title of the 
     * JFrame. It also sets the instance fields to a reasonable default.
     */
    public SnapShopGUI() {
        
        super("TCSS 305 SnapShop");
        myFileChooser = new JFileChooser();
        myLabelForViewingIcon = new JLabel();
        myButtonArray = new ArrayList<>();
        myPixelStack = new ArrayDeque<>();
        myOpen = "Open...";
        myUndo = "Undo";
    }

    /**
     * The main method of this class called by SnapShopMain to start creating the 
     * JFrame and it's implementation. It does this by making a call to setLayout() 
     * and then packing the screen size to fit around the various components.
     */
    public void start() {  
        
        setLocation(SCREEN_WIDTH / 4 - getWidth() / 4, SCREEN_HEIGHT / 4 - getHeight() / 4);
        setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        
        
        setLayout();
        pack();
    }
    
    /** This method is called by start() and sets the basic layout of the JFrame as well as 
     * calling helper methods to create north and south panels.
     */
    private void setLayout() {
        
        setLayout(new BorderLayout());
        final JPanel centerPanel = new JPanel(new GridBagLayout());
        add(centerPanel, BorderLayout.CENTER);
        
        centerPanel.add(myLabelForViewingIcon);
        myLabelForViewingIcon.setVisible(true);
        
        createNorthPanel();
        createSouthPanel();
        
        setVisible(true);
    }
    
    /** This method creates and adds buttons to the north panel by making calls to 
     *  a helper method called createNorthButtons().
     */
    private void createNorthPanel() {
            
        final JPanel northPanel = new JPanel();
            
        northPanel.add(createNorthButtons(new EdgeDetectFilter()));
        northPanel.add(createNorthButtons(new EdgeHighlightFilter()));
        northPanel.add(createNorthButtons(new FlipHorizontalFilter()));
        northPanel.add(createNorthButtons(new FlipVerticalFilter()));
        northPanel.add(createNorthButtons(new GrayscaleFilter()));
        northPanel.add(createNorthButtons(new SharpenFilter()));
        northPanel.add(createNorthButtons(new SoftenFilter()));
            
        add(northPanel, BorderLayout.NORTH);
    }

    /**
     * Creates a single button that applies a filter object's description as it's text. 
     * it then disables the initial state of the button, and it adds an ActionListener via an 
     * anonymous class. The code in the anonymous class for the ActionListener allows the 
     * button to apply a filter to the current image when pressed. This method also pushes 
     * Pixel data onto a stack to use for undoing changes made to the image, and then adds the 
     * finished JButton to an ArrayList used for disabling and enabling the buttons. 
     * 
     * @param theFilterObject The object used to apply various filters to a PixelImage
     * @return the filter button that is used to line the north panel
     */
    private JButton createNorthButtons(final Filter theFilterObject) {
        
        final JButton button = new JButton(theFilterObject.getDescription());
        button.setEnabled(false);
        
        button.addActionListener(new ActionListener() {
           
            /** This method applies filters to an image when a button is clicked as well as 
             *  activating the "Undo" button.
             * @param theAction the action that corresponds to what the user does*/
            public void actionPerformed(final ActionEvent theAction) {
                
                theFilterObject.filter(myImage);
                myPixelStack.push(myImage.getPixelData());
                myLabelForViewingIcon.setIcon(new ImageIcon(myImage));
                switchUndoButtonStatus(true);
            }
        });
        myButtonArray.add(button);
        return button;
    }
    
    /** This method creates and adds buttons to the south panel. It sets the button's text to 
     *  a string held in the southButtonName array, and then enables or disables the button's 
     *  status when appropriate. ActionListeners are then added to the buttons via an Inner 
     *  Class called Listener, and Action Commands are assigned to the buttons so the Listener 
     *  knows what to do when the buttons are pressed. Lastly the south buttons are added to 
     *  myButtonArray, joining the north buttons to be later used for disabling and enabling.
     */
    private void createSouthPanel() {
        
        final JPanel southPanel = new JPanel();
        
        final String[] southButtonName = {"Close Image", myUndo, "Save As...", myOpen};
        
        final JButton[] southButtons = new JButton[southButtonName.length];
        
        for (int i = 0; i < southButtonName.length; i++) {
            
            southButtons[i] = new JButton(southButtonName[i]);
        }
        
        for (int i = southButtons.length - 1; i >= 0; i--) {
            
            southPanel.add(southButtons[i]);
            southButtons[i].setEnabled(false);
            
            if (i == southButtons.length - 1) {
                
                southButtons[i].setEnabled(true);
            }
            
            southButtons[i].addActionListener(new Listener());
            southButtons[i].setActionCommand("command " + i);
            myButtonArray.add(0, southButtons[i]);
        }
        
        add(southPanel, BorderLayout.SOUTH);
    }
     
    /**
     * This method opens an image, as well as showing error message dialogs if an invalid file 
     * was chosen. It also pushes the selected images Pixel data onto a stack for later use in 
     * the undo() method. openFile() enables all of the buttons except the "undo" button when 
     * the file is opened. Lastly, when a file is opened myPixelStack is cleared of all 
     * previous data so that the undo button does not try and revert different pictures to 
     * states of one another.
     */
    private void openFile() {
 
        myPixelStack.clear();    
        final int openValue = myFileChooser.showOpenDialog(myFileChooser);
              
        if (openValue == JFileChooser.APPROVE_OPTION) {

            final File retrievedFile = myFileChooser.getSelectedFile();
                  
                  
            try {
                      
                myImage = PixelImage.load(retrievedFile);
                      
            } catch (final IOException noActions) {
                      
                JOptionPane.showMessageDialog(this, "File did not contain a valid image: " 
                                                                             + retrievedFile);
            }
            if (myImage == null) {
                      
                JOptionPane.showMessageDialog(this, "File was null: " + retrievedFile); 
            } else {
                      
                myPixelStack.push(myImage.getPixelData());
                myLabelForViewingIcon.setIcon(new ImageIcon(myImage));
                      
                for (int i = 0; i < myButtonArray.size(); i++) {

                    myButtonArray.get(i).setEnabled(true); 
                }
                switchUndoButtonStatus(false);
            }      
        } 
        pack();
    }
          
    /**
     * This method saves the currently loaded image to the directory of the user's choice.
     */
    private void saveFile() {
              
        final int saveValue = myFileChooser.showSaveDialog(myFileChooser);
                  
        if (saveValue == JFileChooser.APPROVE_OPTION) {
                      
            final File savedFile = myFileChooser.getSelectedFile();
            try {
                          
                myImage.save(savedFile);
                          
            } catch (final IOException noActions) {
                          
                JOptionPane.showMessageDialog(this, "An error has occurred: " + savedFile);
            }
        }
    }
    
    /**
     * This method reverts the look of an applied filter to the previous images state. 
     * Whenever an image is loaded or a filter is applied, the image's pixel data is pushed 
     * onto a stack. After A filter is applied, the "Undo" button becomes active, and when 
     * pressed, the previously pushed pixel data is popped off and re-applied to the current 
     * image. Also when a new image is opened the stack is cleared so the process can restart.
     */
    private void undo() {
              
        myPixelStack.pop();
        myImage.setPixelData(myPixelStack.peek());
        myLabelForViewingIcon.setIcon(new ImageIcon(myImage));
        myLabelForViewingIcon.setVisible(true);
            
        if (myPixelStack.size() <= 1) {
            
            switchUndoButtonStatus(false);
        }
    }
    
    /**
     * This is a helper method that switches the "Undo" button to be either false for 
     * inactive, or true for active. 
     * Used in the createNorthButtons(), openFile(), and undo() methods.
     * 
     * @param theStatus which is the true or false state being applied to the "Undo" button
     */
    private void switchUndoButtonStatus(final boolean theStatus) {
        
        for (int i = 0; i < myButtonArray.size(); i++) {

            if (myUndo.equals(myButtonArray.get(i).getText())) {
                
                myButtonArray.get(i).setEnabled(theStatus);
            }
        }
    }
          
    /**
     * This method restarts the GUI's interface to it's original state with no picture being 
     * shown and only the "Open.." button activated.
     */
    private void closeImage() {
              
        myLabelForViewingIcon.setIcon(null);
        for (int i = 0; i < myButtonArray.size(); i++) {
                  
            myButtonArray.get(i).setEnabled(false); 
            
            if (myOpen.equals(myButtonArray.get(i).getText())) {
                
                myButtonArray.get(i).setEnabled(true);
            }
        }
        
        pack();
    }
          
    /**
     * The Listener class extends ActionListener and allows ActionListeners to be applied to 
     * components. This class also keeps track of it's components actions implementing 
     * appropriate behaviors when certain JButtons are pressed. 
     * 
     * @author Brian Dennis
     * @version 6 February 2014
     */
    class Listener implements ActionListener {
              
        /**
         * This method calls various helper methods depending on what action is performed.
         * 
         * @param theAction the action that corresponds to what the user does
         */
        public void actionPerformed(final ActionEvent theAction) {
               
            myFileChooser.setCurrentDirectory(myFileChooser.getCurrentDirectory());
            final String action = theAction.getActionCommand();
                    
            if ("command 3".equals(action)) {
                openFile();
            }
                    
            if ("command 2".equals(action)) {
                saveFile();
            }
                    
            if ("command 1".equals(action)) { 
                undo();
            }
                    
            if ("command 0".equals(action)) {
                closeImage();
            }
        }
    } 
}